local addon = ActionDurationReminder --Addon#Addon

---
--@type CoreSavedVars
local coreSavedVarsDefaults = {
    coreMultipleTargetTracking = true,
    coreSecondsBeforeFade = 5,
    coreMinimumDurationSeconds = 3,
}

local SPECIAL_ABILITY_IDS = {
    CRYSTAL_FRAGMENT_PROC_EFFECT = 46327, -- detect this proc triggered randomly
    CRYSTAL_FRAGMENT_EFFECT = 47570, -- ignore crystal stunning to avoid messing up with proc
    CRYSTAL_FRAGMENT = 47569, -- stop proc duration as soon as casting ability
    MAJOR_EXPEDITION_OF_CRIPPLING = 37915, -- ignore its buggy time info(<Crippling Grasp> + <Path of Darkness>)
    PURIFYING_LIGHT_TICK = 68582, -- ignore purifying light healing after main duration
    LIGHTINING_SPLASH = 30291, -- ignore this (n+1)s redundant effect 
}

---
--@type DebugConfig
local debugLevels = {
    effect = 0,
    action = 0,
    target = 0,
}

local function debug(switch, ...)
    if switch then
        d(...)
    end
end

local lastAction = nil --adr.model.Action#Action

---
--@return #CoreSavedVars
local function SavedVars()
    return addon.SavedVars()
end

---
--@type WeaponPairInfo
local weaponPairInfo = {
    index = GetActiveWeaponPairInfo()--#number
    ,
    ultimate = false,
}

---
--@module Core
--@extends Addon#Class
local module = addon.AddClass(
    "adr.Core"
    ,
    ---
    -- @callof #Core
    -- @param #Core self
    -- @return #Core
    function(self)
        self.idActionMap = {}--#map<#number,adr.model.Action#Action>
        self.timeActionMap = {}--#map<#number,adr.model.Action#Action>
        self.actionQueue = {} --#list<adr.model.Action#Action>
        self.weaponPairInfo = weaponPairInfo--#WeaponPairInfo
        EVENT_MANAGER:RegisterForEvent(addon.name, EVENT_ACTION_SLOT_ABILITY_USED, function(...) self:OnSlotAbilityUsed(...) end)
        EVENT_MANAGER:RegisterForUpdate(addon.name, 100, function(...) self:OnUpdate(...) end)
        EVENT_MANAGER:RegisterForEvent(addon.name, EVENT_ACTION_SLOTS_FULL_UPDATE, function(...) self:OnSlotsFullUpdate(...) end)
        EVENT_MANAGER:RegisterForEvent(addon.name, EVENT_EFFECT_CHANGED, function(...) self:OnEffectChanged(...) end)
        EVENT_MANAGER:AddFilterForEvent(addon.name, EVENT_EFFECT_CHANGED, REGISTER_FILTER_SOURCE_COMBAT_UNIT_TYPE, COMBAT_UNIT_TYPE_PLAYER)
        EVENT_MANAGER:RegisterForEvent(addon.name, EVENT_RETICLE_TARGET_CHANGED, function(...) self:OnTargetChanged(...) end)
    end
    ,
    ---
    --@param #Core class
    function(class)
        addon.AddExtension("adr.Settings:BuildDefaults", function(settings, defaults)
            for key, var in pairs(coreSavedVarsDefaults) do
                defaults[key] = var
            end
        end)
        addon.AddExtension("adr.Settings:BuildOptionsData", function(settings, optionsData)
            local gettext = LibStub("LibTextDict")("ActionDurationReminder").gettext
            optionsData[#optionsData+1] = {
                type = "checkbox",
                name = gettext("Multiple Target Tracking"),
                getFunc = function() return SavedVars().coreMultipleTargetTracking end,
                setFunc = function(value) SavedVars().coreMultipleTargetTracking = value end,
                width = "full",
                default = coreSavedVarsDefaults.coreMultipleTargetTracking,
            }
            optionsData[#optionsData+1] = {
                type = "slider",
                name = gettext("Seconds to Keep Timers After Timeout"),
                min = 0, max = 10, step = 1,
                getFunc = function() return SavedVars().coreSecondsBeforeFade end,
                setFunc = function(value) SavedVars().coreSecondsBeforeFade = value end,
                width = "full",
                default = coreSavedVarsDefaults.coreSecondsBeforeFade,
            }
            optionsData[#optionsData+1] = {
                type = "slider",
                name = gettext("Seconds of Ignorable Short Timers"),
                min = 2, max = 5, step = 0.5,
                getFunc = function() return SavedVars().coreMinimumDurationSeconds end,
                setFunc = function(value) SavedVars().coreMinimumDurationSeconds = value end,
                width = "full",
                default = coreSavedVarsDefaults.coreMinimumDurationSeconds,
            }
        end)
        addon.AddExtension("addon:OnStarted",function(addon) class() end)
    end
) --#Core

---
--@param #Core self
--@param adr.model.Effect#Effect effect
function module:FindActionByNewEffect(effect)
    for i = 1,#self.actionQueue do
        local action = self.actionQueue[i]
        if action:MatchNewEffect(effect)then
            debug(debugLevels.action>0 , '[F]found one by new match:'..action.ability.name..'@'..(action.startTime/1000))
            return action
        end
    end
    debug(debugLevels.action>0 , '[?]not found in '..#self.actionQueue..' actions, last:'..(lastAction and (lastAction.ability.name) or 'nil'))
    return nil
end

---
--@param #Core self
--@param adr.model.Effect#Effect effect
--@param #boolean oldFirst
function module:FindActionByOldEffect(effect, oldFirst)
    local from = oldFirst and #self.actionQueue or 1
    local to = oldFirst and  1 or #self.actionQueue
    local step = oldFirst and -1 or 1
    -- 1. find that can
    for i = from,to,step do
        local action = self.actionQueue[i] --adr.model.Action#Action
        if action:MatchOldEffect(effect) then
            if not oldFirst and action.overideAction then action = action.overideAction end
            debug(debugLevels.action>0 , '[F]found one by old match:'..action.ability.name..'@'..(action.startTime/1000))
            return action
        end
    end
    for i = from,to,step do
        local action = self.actionQueue[i] --adr.model.Action#Action
        if action:MatchNewEffect(effect)then
            debug(debugLevels.action>0 , '[F]found one by new match:'..action.ability.name..'@'..(action.startTime/1000))
            return action
        end
    end
    debug(debugLevels.action>0 , '[?]not found in '..#self.actionQueue..' actions, last:'..(lastAction and (lastAction.ability.name) or 'nil'))
    return nil
end

---
--@param #Core self
-- @param #number eventCode
-- @param #number changeType
-- @param #number effectSlot
-- @param #string effectName
-- @param #string unitTag
-- @param #number beginTime
-- @param #number endTime
-- @param #number stackCount
-- @param #string iconName
-- @param #string buffType
-- @param #number effectType
-- @param #number abilityType
-- @param #number statusEffectType
-- @param #string unitName
-- @param #number unitId
-- @param #number abilityId
--
function module:OnEffectChanged(eventCode, changeType, effectSlot, effectName, unitTag, startTimeSec, endTimeSec,
    stackCount, iconName, buffType, effectType, abilityType, statusEffectType, unitName, unitId, effectAbilityId)
    local now = GetGameTimeMilliseconds()
    debug(debugLevels.effect>0, '['..(changeType==1 and '+' or changeType==2 and '-' or changeType==3 and '=' or '?')..
        (stackCount>0 and stackCount or'')..']'..effectName..'('..effectAbilityId..')@'..(startTimeSec>0 and startTimeSec
        or ('n'..now/1000))..'<'..(endTimeSec-startTimeSec)..'>['..iconName..'] for '..
        (unitTag~='' and unitTag or 'none')..'('..unitId..')')
    -- 0. prepare
    if effectAbilityId == SPECIAL_ABILITY_IDS.PURIFYING_LIGHT_TICK then return end
    if effectAbilityId == SPECIAL_ABILITY_IDS.CRYSTAL_FRAGMENT_EFFECT then return end
    if effectAbilityId == SPECIAL_ABILITY_IDS.MAJOR_EXPEDITION_OF_CRIPPLING then return end
    if effectAbilityId == SPECIAL_ABILITY_IDS.LIGHTINING_SPLASH then return end
    if unitTag and string.find(unitTag, 'group') then return end -- ignore effects on group members especially those same as player
    local startTime =  math.floor(startTimeSec * 1000)
    local endTime =  math.floor(endTimeSec * 1000)
    local duration = endTime-startTime
    if duration > 100000 then return end -- ignore effects that last longer than 100 seconds
    if lastAction and not lastAction.flags.isGround and startTime and startTime  - (lastAction.startTime +lastAction.castTime)> 2000 then
        debug(debugLevels.action>0, '[w] wipe last by time')
        lastAction = nil
    end
    local Ability = addon.GetClass("adr.model.Ability") --adr.model.Ability#Ability
    local ability = Ability(effectAbilityId,effectName,iconName)
    local Effect = addon.GetClass("adr.model.Effect") --adr.model.Effect#Effect
    local effect = Effect(ability, unitTag, unitId, startTime, endTime)
    -- 1. stack
    if stackCount > 0 then -- e.g. relentless focus
        local action = self:FindActionByNewEffect(effect)--adr.model.Action#Action
        if not action then return end
        if action.duration>0 then -- stackable actions with duration should ignore eso buggy effect time e.g. 20s Relentless Focus
            effect.startTime = action.startTime
            effect.duration = action.duration
            effect.endTime = action.endTime
        end
        if changeType == EFFECT_RESULT_FADED then
            action.stackCount = 0
            action:PurgeEffect(effect)
        else
            action.stackCount = stackCount
            action:SaveEffect(effect)
            self:SaveAction(action)
        end
        return
    end
    if duration>0 and duration < SavedVars().coreMinimumDurationSeconds*1000 +100 then return end
    -- 2. gain
    if changeType == EFFECT_RESULT_GAINED then
        -- 2.x crystal fragment proc show at its own skill, not last action
        if effectAbilityId == SPECIAL_ABILITY_IDS.CRYSTAL_FRAGMENT_PROC_EFFECT then -- proc happens
            local crystalSlotNum = nil
            for slotNum = 3,7 do
                local slotBoundId = GetSlotBoundId(slotNum)
                if SPECIAL_ABILITY_IDS.CRYSTAL_FRAGMENT== slotBoundId
                    or string.find(effect.ability.name,zo_strformat("<<1>>", GetSlotName(slotNum)),1,true)
                then
                    crystalSlotNum = slotNum
                    SPECIAL_ABILITY_IDS.CRYSTAL_FRAGMENT= slotBoundId -- this ability id depends on level
                    break
                end
            end
            if crystalSlotNum then
                local Action = addon.GetClass("adr.model.Action") -- adr.model.Action#Action
                local action = Action(crystalSlotNum, weaponPairInfo.index, weaponPairInfo.ultimate)
                action:SaveEffect(effect)
                self:SaveAction(action)
            end
            return
        end
        -- 2.x normal situation
        local action = self:FindActionByNewEffect(effect)--adr.model.Action#Action
        if action then
            action:SaveEffect(effect)
            self:SaveAction(action)
            return
        end
        debug(debugLevels.effect>0, 'effect action not found')
        return
    end
    -- 3. update
    if changeType == EFFECT_RESULT_UPDATED then
        local action = self:FindActionByOldEffect(effect)--adr.model.Action#Action
        if action then
            action:SaveEffect(effect)
            self:SaveAction(action)
            return
        end
        return
    end
    -- 4. fade
    if changeType == EFFECT_RESULT_FADED then
        local action = self:FindActionByOldEffect(effect, true)--adr.model.Action#Action
        if action then
            action:PurgeEffect(effect)
            if action:EndTime()<=now then
                debug(debugLevels.action>0, '[P]'..action.ability.name..'('..action.ability.id..')['..
                    action.ability.icon..']@'..(action.startTime/1000))
            end
            --4.x Crystal fragment doesn't need blink
            if action.ability.id == SPECIAL_ABILITY_IDS.CRYSTAL_FRAGMENT then
                self.idActionMap[action.ability.id] = nil
            end
            return
        end
        return
    end
end

---
--@param #Core self
--@param #number eventCode
--@param #number slotNum
function module:OnSlotAbilityUsed(eventCode, slotNum)
    -- 1. filter other actions
    if slotNum < 3 or slotNum > 8 then return end

    -- 2. create action model
    local Action = addon.GetClass('adr.model.Action') --adr.model.Action#Action
    local action = Action(slotNum, weaponPairInfo.index, weaponPairInfo.ultimate)
    self:QueueAction(action)

    -- 3. special activity performed to turn off
    if action.ability.id == SPECIAL_ABILITY_IDS.CRYSTAL_FRAGMENT then
        self:RemoveAction(action) -- crystal fragment timer end when used
        return
    end
end

---
--@param #Core self
--@param #number eventCode
--@param #boolean isHotbarSwap
function module:OnSlotsFullUpdate(eventCode, isHotbarSwap)
    if not isHotbarSwap then return end

    local newPairIndex,_ = GetActiveWeaponPairInfo()
    if newPairIndex ~= weaponPairInfo.index then
        weaponPairInfo.index = newPairIndex
        weaponPairInfo.ultimate = false
    else
        weaponPairInfo.ultimate = not weaponPairInfo.ultimate
    end
end

---
--@param #Core self
function module:OnTargetChanged()
    if not SavedVars().coreMultipleTargetTracking then return end
    if not DoesUnitExist('reticleover') then return end
    -- 1. remove all others-effect actions from self.idActionMap
    for key,var in pairs(self.idActionMap) do
        local action = var--adr.model.Action#Action
        if not action.flags.isGround and not action.flags.isArea and not action:IsOnPlayer() then
            self.idActionMap[key] = nil
            debug(debugLevels.target>0, '[C]'..action.ability.name..'('..action.ability.id..')@'..(action:StartTime()/1000)..'<'..(action:Duration()/1000)..'>')
        end
    end
    -- 2. scan all matched buffs
    local numBuffs = GetNumBuffs('reticleover')
    local Ability = addon.GetClass("adr.model.Ability")--adr.model.Ability#Ability
    local Effect = addon.GetClass("adr.model.Effect")--adr.model.Effect#Effect
    local numRestored = 0
    for i = 1, numBuffs do
        local effectName, startSec, endSec, _, stackCount, icon, _, effectType, abilityType, _, abilityId,_,castByPlayer = GetUnitBuffInfo('reticleover', i)
        if castByPlayer then
            local startTime =  math.floor(startSec * 1000)
            local action = self.timeActionMap[startTime]
            if action then
                local ability = Ability(abilityId,effectName, icon)
                local effect = Effect(ability, 'none', 0, startTime, startTime) -- only for match
                if action:MatchOldEffect(effect) then
                    self.idActionMap[action.ability.id] = action
                    numRestored = numRestored+1
                    debug(debugLevels.target>0, '[V]'..action.ability.name..'('..action.ability.id..')@'..(action:StartTime()/1000)..'<'..(action:Duration()/1000)..'>')
                else
                    debug(debugLevels.target>0, '[X]'..action.ability.name..'('..action.ability.id..')@'..(action:StartTime()/1000)..'<'..(action:Duration()/1000)..'>')
                    debug(debugLevels.target>0, '[?]'..effectName..'('..abilityId..')@'..startSec..'<'..(endSec-startSec)..'> action not verified')
                end
            else
                debug(debugLevels.target>0, '[?]'..effectName..'('..abilityId..')@'..startSec..'<'..(endSec-startSec)..'> action not found')
            end
        end

    end
end

---
--@param #Core self
function module:OnUpdate()
    self:RefineActions()
    self:CallExtension("Update")
end

---
--@param #Core self
function module:RefineActions()
    local now = GetGameTimeMilliseconds()
    local endLimit = now - SavedVars().coreSecondsBeforeFade * 1000
    for key,var in pairs(self.idActionMap) do
        local action = var -- adr.model.Action#Action
        if action:EndTime() < endLimit then
            self:RemoveAction(action)
        end
    end
    endLimit = endLimit - 3000 -- timeActionMap remains a little longer to be found by further effect
    for key,var in pairs(self.timeActionMap) do
        local action = var -- adr.model.Action#Action
        if action:EndTime() < endLimit then
            self.timeActionMap[key] = nil
            debug(debugLevels.action>0, '[d]'..action.ability.name..'('..action.ability.id..')['..
                action.ability.icon..']@'..(action.startTime/1000)..'<'..action:Duration()..'>')
        end
    end
end

---
--@param #Core self
--@param adr.model.Action#Action action
function module:RemoveAction(action)
    self.idActionMap[action.ability.id] = nil
    debug(debugLevels.action>0, '[R]'..action.ability.name..'('..action.ability.id..')['..
        action.ability.icon..']@'..(action.startTime/1000))
end



---
--@param #Core self
--@param adr.model.Action#Action action
function module:QueueAction(action)
    debug(debugLevels.action>0, '[a]'..action.ability.name..'('..action.ability.id..')['..
        action.ability.icon..']@'..(action.startTime/1000)..'+'..(action.castTime/1000)..'++'..(GetLatency()/1000))
    lastAction = action
    local newQueue = {} --#list<adr.model.Action#Action>
    newQueue[1] = action
    for i = 1,#self.actionQueue do
        local a = self.actionQueue[i]
        if a.ability.id == action.ability.id then a.overideAction = action end
        if a:EndTime()+SavedVars().coreSecondsBeforeFade*1000 > action.startTime then
            newQueue[1+#newQueue] = a
        end
    end
    self.actionQueue = newQueue
end

---
--@param #Core self
--@param adr.model.Action#Action action
function module:SaveAction(action)
    self.idActionMap[action.ability.id] = action
    for key, var in pairs(action.effectMap) do
        local effect = var --adr.model.Effect#Effect
        self.timeActionMap[effect.startTime] = action
    end
    local len = function(t)
        local count = 0
        for key, var in pairs(t) do
            count=count+1
        end
        return count
    end
    debug(debugLevels.action>0, '[s]idActionMap('..len(self.idActionMap)..'),timeActionMap('..len(self.timeActionMap)..')')
end

--========================================
return module
