local addon = ActionDurationReminder

---
--@return adr.ui.Bar#BarSavedVars
local function SavedVars()
    return addon.SavedVars()
end

---
--@module ResourceBarPatch
--@extends Addon#Class
local module = addon.AddClass(
    "adr.patch.ResourceBarPatch"
    ,
    ---
    -- @callof #ResourceBarPatch
    -- @param #ResourceBarPatch self
    -- @return #ResourceBarPatch
    function(self)
        self.hudInfo = {}
        self.newRectReady = false
        self.shifted = false
        addon.AddExtension("adr.Core:Update",function(core)
            self:OnCoreUpdate(core)
        end)
    end
    ,
    ---
    --@param #ResourceBarPatch class
    function(class)
        addon.AddExtension("addon:OnStarted",function(addon)
            class()
        end)
    end
) --#ResourceBarPatch

---
--@param #ResourceBarPatch self
--@param adr.Core#Core core
function module:OnCoreUpdate(core)
    -- 1. check if we have done
    local shiftBarVisible = false
    if SavedVars().barShowShift then
        for key, var in pairs(core.idActionMap) do
            local action = var --adr.model.Action#Action
            if action.flags.shifted then
                shiftBarVisible = true
                break
            end
        end
    end
    if not shiftBarVisible and not self.shifted then return end
    if shiftBarVisible and self.shifted then return end
    -- 2. collect info
    local gap = 5
    local hudTable = {
        hb = ZO_PlayerAttributeHealth,
        mb = ZO_PlayerAttributeMagicka,
        sb = ZO_PlayerAttributeStamina,
    }
    -- 2.1 record their original info
    for name,hud in pairs(hudTable) do
        if not self.hudInfo[name] then
            local _, point, relativeTo, relativePoint, offsetX, offsetY = hud:GetAnchor(0)
            local bottom = hud:GetBottom()
            self.hudInfo[name] = {
                point = point,
                relativeTo = relativeTo,
                relativePoint = relativePoint,
                offsetX = offsetX,
                offsetY = offsetY,
                bottom = bottom,
            }
        end
    end

    -- 2.2 compute newRect
    if not self.newRectReady then
        local rectList = {}
        local slot3 = ZO_ActionBar_GetButton(3).slot
        local slot7 = ZO_ActionBar_GetButton(7).slot
        local shiftY = -50 - gap
        local offsetX = SavedVars().barShiftOffsetX
        local offsetY = SavedVars().barShiftOffsetY
        rectList[#rectList+1]={
            top = slot3:GetTop() + shiftY + offsetY,
            bottom = slot3:GetBottom() + shiftY + offsetY,
            left = slot3:GetLeft() + offsetX,
            right = slot7:GetRight() + offsetX,
        }
        local nameList = {'hb','mb','sb'}
        table.sort(nameList, function(name1, name2)
            return hudTable[name1]:GetBottom()> hudTable[name2]:GetBottom()
        end)
        for index = 1,#nameList do
            local name = nameList[index]
            local hud = hudTable[name]
            local hudRect, newOffsetY, newBottom = self:ComputeNewOffsetY(rectList, name, hud, gap)
            rectList[#rectList+1]= hudRect
            self.hudInfo[name].newRect = hudRect
        end
        self.newRectReady = true
    end
    -- 3. check bottom to set anchor
    for name,hud in pairs(hudTable) do
        local info = self.hudInfo[name]
        local bottom = shiftBarVisible and info.newRect.bottom or info.bottom
        if bottom ~= hud:GetBottom() then
            hud:ClearAnchors()
            if shiftBarVisible then
                hud:SetAnchor(TOPLEFT, GuiRoot, TOPLEFT, info.newRect.left, info.newRect.top)
                hud:SetAnchor(BOTTOMRIGHT, GuiRoot, TOPLEFT, info.newRect.right, info.newRect.bottom)
            else
                hud:SetAnchor(info.point, info.relativeTo, info.relativePoint, info.offsetX, info.offsetY)
            end
        end
    end
    -- 4. reset data
    self.shifted = shiftBarVisible
    if not shiftBarVisible then
        self.hudInfo = {}
        self.newRectReady = false
    end
end

---
--@param #ResourceBarPatch self
--@param #list<#table> rectList
--@param #string name
--@param #number gap
function module:ComputeNewOffsetY(rectList, name, hud, gap)
    local hudRect = {
        left = hud:GetLeft(),
        right = hud:GetRight(),
        top = hud:GetTop(), -- bar have out margin
        bottom = hud:GetBottom(), -- bar have out margin
    }
    local delta = 0
    for _,rect in pairs(rectList) do
        delta = delta + self:MoveUp(hudRect, rect, gap)
    end
    return hudRect, self.hudInfo[name].offsetY + delta, self.hudInfo[name].bottom + delta
end

---
--@param #ResourceBarPatch self
--@param #table movingRect
--@param #table rect
--@param #number gap
function module:MoveUp(movingRect, rect, gap)
    if movingRect.left > rect.right or movingRect.right < rect.left then return 0 end
    if movingRect.bottom + gap < rect.top or movingRect.top - gap > rect.bottom then return 0 end
    local delta = rect.top - gap - movingRect.bottom
    movingRect.top = movingRect.top + delta
    movingRect.bottom = movingRect.bottom + delta
    return delta
end

return module
