local addon = ActionDurationReminder

---
--@type AlertSavedVars
local alertSavedVarsDefaults = {
    alertEnabled = false,
    alertPlaySound = true,
    alertSoundIndex = 266,
    alertAheadSeconds = 1,
    alertKeepSeconds = 2,
    alertIconSize = 50,
    alertFontName = "BOLD_FONT",
    alertFontSize = 32,
    alertOffsetX = 0,
    alertOffsetY = 0,
}

local soundChoices = {}
for k,v in pairs(SOUNDS) do
    soundChoices[#soundChoices+1]=k
end
table.sort(soundChoices)

---
--@return #AlertSavedVars
local function SavedVars()
    return addon.SavedVars()
end

local alert --#Alert

---
--@module Alert
--@extends Addon#Class
local module = addon.AddClass(
    "adr.ui.Alert"
    ,
    ---
    -- @callof #Alert
    -- @param #Alert self
    -- @return #Alert
    function(self)
        self.controlPool = nil
        self.showList = {}
        self.frame = nil --TopLevelWindow#TopLevelWindow
        addon.AddExtension("adr.Core:Update",function(core)
            self:OnCoreUpdate(core)
        end)
    end
    ,
    ---
    --@param #Alert class
    function(class)
        addon.AddExtension("adr.Settings:BuildDefaults", function(settings, defaults)
            for key, var in pairs(alertSavedVarsDefaults) do
                defaults[key] = var
            end
        end)
        addon.AddExtension("adr.Settings:BuildOptionsData", function(settings, optionsData)
            local gettext = LibStub("LibTextDict")("ActionDurationReminder").gettext
            optionsData[#optionsData+1] = {
                type = "header",
                name = gettext("Popup Alert"),
                width = "full",
            }
            optionsData[#optionsData+1] = {
                type = "checkbox",
                name = gettext("Enabled"),
                getFunc = function() return SavedVars().alertEnabled end,
                setFunc = function(value) SavedVars().alertEnabled = value end,
                width = "full",
                default = alertSavedVarsDefaults.alertEnabled,
            }
            optionsData[#optionsData+1] = {
                type = "button",
                name = gettext("Move Alert Frame"),
                func = function()
                    SCENE_MANAGER:Hide("gameMenuInGame")
                    alert:OpenAlertFrame()
                    zo_callLater(function()
                        SetGameCameraUIMode(true)
                    end, 10)
                end,
                width = "full",
                disabled = function() return not SavedVars().alertEnabled end,
            }
            optionsData[#optionsData+1] = {
                type = "checkbox",
                name = gettext("Sound Enabled"),
                getFunc = function() return SavedVars().alertPlaySound end,
                setFunc = function(value) SavedVars().alertPlaySound = value end,
                disabled = function() return not SavedVars().alertEnabled end,
                width = "full",
                default = alertSavedVarsDefaults.alertPlaySound,
            }
            optionsData[#optionsData+1] = {
                type = "slider",
                name = gettext("Sound Select Index"),
                --tooltip = "",
                min = 1, max = #soundChoices, step = 1,
                getFunc = function() return SavedVars().alertSoundIndex end,
                setFunc = function(value) SavedVars().alertSoundIndex = value; PlaySound(SOUNDS[soundChoices[SavedVars().alertSoundIndex]]) end,
                width = "full",
                disabled = function() return not SavedVars().alertPlaySound or not SavedVars().alertEnabled end,
                default = alertSavedVarsDefaults.alertSoundIndex,
            }
            optionsData[#optionsData+1] = {
                type = "button",
                name = gettext("Sound Test"),
                func = function()
                    PlaySound(SOUNDS[soundChoices[SavedVars().alertSoundIndex]])
                end,
                width = "full",
                disabled = function() return not SavedVars().alertPlaySound or not SavedVars().alertEnabled end,
            }
            optionsData[#optionsData+1] = {
                type = "slider",
                name = gettext("Seconds to Show Before End"),
                --tooltip = "",
                min = 0, max = 3, step = 0.5,
                getFunc = function() return SavedVars().alertAheadSeconds end,
                setFunc = function(value) SavedVars().alertAheadSeconds = value end,
                width = "full",
                disabled = function() return not SavedVars().alertEnabled end,
                default = alertSavedVarsDefaults.alertAheadSeconds,
            }
            optionsData[#optionsData+1] = {
                type = "slider",
                name = gettext("Seconds to Show"),
                --tooltip = "",
                min = 1, max = 5, step = 0.5,
                getFunc = function() return SavedVars().alertKeepSeconds end,
                setFunc = function(value) SavedVars().alertKeepSeconds = value end,
                width = "full",
                disabled = function() return not SavedVars().alertEnabled end,
                default = alertSavedVarsDefaults.alertKeepSeconds,
            }
            optionsData[#optionsData+1] = {
                type = "dropdown",
                name = gettext("Font Name"),
                choices = {"MEDIUM_FONT", "BOLD_FONT", "CHAT_FONT", "ANTIQUE_FONT", "HANDWRITTEN_FONT", "STONE_TABLET_FONT", "GAMEPAD_MEDIUM_FONT", "GAMEPAD_BOLD_FONT"},
                getFunc = function() return SavedVars().alertFontName end,
                setFunc = function(value) SavedVars().alertFontName = value end,
                disabled = function() return not SavedVars().alertEnabled end,
                width = "full",
                default = alertSavedVarsDefaults.alertFontName,
            }
            optionsData[#optionsData+1] = {
                type = "slider",
                name = gettext("Font Size"),
                --tooltip = "",
                min = 18, max = 48, step = 2,
                getFunc = function() return SavedVars().alertFontSize end,
                setFunc = function(value) SavedVars().alertFontSize = value end,
                disabled = function() return not SavedVars().alertEnabled end,
                width = "full",
                default = alertSavedVarsDefaults.alertFontSize,
            }
        end)
        addon.AddExtension("addon:OnStarted",function(addon)
            alert=class()
        end)
    end
) --#Alert


---
--@param #Alert self
--@param adr.model.Ability#Ability ability
function module:Alert(ability)
    if SavedVars().alertPlaySound then PlaySound(SOUNDS[soundChoices[SavedVars().alertSoundIndex]]) end

    if not self.controlPool then
        self.controlPool = ZO_ObjectPool:New(function(pool)
            local control = WINDOW_MANAGER:CreateTopLevelWindow()
            local icon = control:CreateControl(nil, CT_TEXTURE)
            icon:SetDrawLayer(DL_CONTROLS)
            icon:SetDimensions(SavedVars().alertIconSize,SavedVars().alertIconSize)
            icon:SetAnchor(LEFT)
            control.icon = icon
            local label = control:CreateControl(nil, CT_LABEL)
            label:SetFont("ZoFontGameMedium")
            label:SetColor(1,1,1)
            label:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
            label:SetAnchor(LEFT, icon, RIGHT, 10, 0)
            label:SetDrawLayer(DL_TEXT)
            control.label = label
            return control
        end,
        function(control)
            control:SetHidden(true)
            control:ClearAnchors()
        end)
    end
    local control,key = self.controlPool:AcquireObject()
    control.label:SetFont("$("..SavedVars().alertFontName..")|"..SavedVars().alertFontSize.."|thick-outline")
    control.label:SetText(zo_strformat('<<C:1>>',ability.name))
    control.icon:SetTexture(ability.icon)
    control:SetAnchor(BOTTOMLEFT, GuiRoot, CENTER, -150 + SavedVars().alertOffsetX, -150 + SavedVars().alertOffsetY)
    control:SetHidden(false)
    local showList = self.showList
    for k,v in pairs(showList) do
        local _, _, _, _, offsetX, offsetY = v:GetAnchor(0)
        v:ClearAnchors()
        v:SetAnchor(BOTTOMLEFT, GuiRoot, CENTER, offsetX, offsetY -SavedVars().alertIconSize-10)
    end
    showList.key = control
    zo_callLater(
        function()
            control:SetHidden(true)
            showList.key = nil
            self.controlPool:ReleaseObject(key)
        end,
        SavedVars().alertKeepSeconds*1000
    )
end

---
--@param #Alert self
--@param adr.Core#Core core
function module:OnCoreUpdate(core)
    local now = GetGameTimeMilliseconds()
    for key,var in pairs(core.idActionMap) do
        local action = var --adr.model.Action#Action
        if SavedVars().alertEnabled and action:EndTime() - SavedVars().alertAheadSeconds *1000 < now and not action.data.alerted then
            action.data.alerted = true
            self:Alert(action.ability)
        end
    end
end

---
--@param #Alert self
function module:OpenAlertFrame()
    if not self.frame then
        self.frame = WINDOW_MANAGER:CreateTopLevelWindow()
        self.frame:SetDimensions(350, 70)
        self.frame:SetMouseEnabled(true)
        self.frame:SetMovable(true)
        self.frame:SetDrawLayer(DL_COUNT)
        self.frame:SetHandler('OnMoveStop', function()
            local left = self.frame:GetLeft()
            local bottom = self.frame:GetBottom()
            local centerX,centerY = GuiRoot:GetCenter()
            SavedVars().alertOffsetX = left - centerX + 150
            SavedVars().alertOffsetY = bottom - centerY + 150
            zo_callLater(function()
                SetGameCameraUIMode(true)
            end, 10)
        end)
        local backdrop = WINDOW_MANAGER:CreateControl(nil, self.frame, CT_BACKDROP) --BackdropControl#BackdropControl
        backdrop:SetAnchor(TOPLEFT)
        backdrop:SetAnchor(BOTTOMRIGHT)
        backdrop:SetCenterColor(0,0,1,0.5)
        backdrop:SetEdgeTexture('',1,1,1,1)
        local label = WINDOW_MANAGER:CreateControl(nil, backdrop, CT_LABEL) --LabelControl#LabelControl
        label:SetFont('$(MEDIUM_FONT)|$(KB_18)|soft-shadow-thin')
        label:SetColor(1,1,1)
        label:SetHorizontalAlignment(TEXT_ALIGN_CENTER)
        label:SetVerticalAlignment(TEXT_ALIGN_CENTER)
        label:SetAnchor(CENTER)
        label:SetDrawLayer(DL_COUNT)
        label:SetDrawLevel(1)
        label:SetText('ADR Alert Frame')
        local labelClose = WINDOW_MANAGER:CreateControl(nil, backdrop, CT_LABEL) --LabelControl#LabelControl
        labelClose:SetFont('$(MEDIUM_FONT)|$(KB_18)|soft-shadow-thin')
        labelClose:SetColor(1,1,1)
        labelClose:SetHorizontalAlignment(TEXT_ALIGN_RIGHT)
        labelClose:SetVerticalAlignment(TEXT_ALIGN_BOTTOM)
        labelClose:SetAnchor(BOTTOMRIGHT, nil, BOTTOMRIGHT, -5, -2)
        labelClose:SetDrawLayer(DL_COUNT)
        labelClose:SetDrawLevel(1)
        labelClose:SetText('[X]')
        labelClose:SetMouseEnabled(true)
        labelClose:SetHandler('OnMouseUp', function() self.frame:SetHidden(true) end)
    end
    self.frame:SetHidden(false)
    self.frame:ClearAnchors()
    self.frame:SetAnchor(BOTTOMLEFT, GuiRoot, CENTER, - 150+SavedVars().alertOffsetX, -150+SavedVars().alertOffsetY)
end


return module
